
-- a star pathfinding algorithm

local Astar = {}

function Astar.setMap(_map)
	Astar.map = _map
end

function Astar.run(start, goal)

	Astar.open		= {}
	Astar.closed	= {}
	
	-- check that map is available
	if Astar.map == nil then
		return false
	end
	
	-- check that the obstacles are set properly
	if type(Astar.obstacles) ~= "table" then
		print("a* error: obstacle check failed!")
		return false
	end
	
	-- check that start and goal are inside the map and not blocked itself
	if (start.x < 1) or (start.x > Astar.map.width) or (start.y < 1) or (start.y > Astar.map.height)
	or (goal.x < 1) or (goal.x > Astar.map.width) or (goal.y < 1) or (goal.y > Astar.map.height) then
		print("a* error: position check failed!")
		return false
	end
	if (tools.isValueInTable(Astar.map.blockingData[goal.y][goal.x], Astar.obstacles)) or (tools.isValueInTable(Astar.map.blockingData[start.y][start.x], Astar.obstacles)) then
		print("a* error: initial blocking check failed!")
		return false
	end
	
	table.insert(Astar.open, start)
	
	start.g = 0
	start.f = start.g + Astar.heuristic(start, goal)
	
	
	while #Astar.open > 0 do
	
		--print("a* open nodes: " .. #Astar.open)
		local current = Astar.getElementWithLowestF()
		--print("a* current node: " .. current.x .. " " .. current.y)
		if Astar.isNodeEqual(current, goal) then
			return Astar.constructPath(current)
		end
		-- otherwise...
		Astar.removeNode(Astar.open, current)
		table.insert(Astar.closed, current)
		
		for _, neighbour in pairs(Astar.getNeighbours(current)) do
			--print("neighbour parent: [" .. neighbour.parent.x .. ":" .. neighbour.parent.y .. "]")
			
			if not Astar.isNodeInList(neighbour, Astar.closed) then
				neighbour.f = neighbour.g + Astar.heuristic(neighbour, goal)
				
				if not Astar.isNodeInList(neighbour, Astar.open) then
					--print("insert node " .. neighbour.x .. " " .. neighbour.y .. " into open list")
					table.insert(Astar.open, neighbour)
				else
					--print("node " .. neighbour.x .. " " .. neighbour.y .. " is already in open list")
					for _, openNeighbour in pairs(Astar.open) do
						--print("check neighbour " .. openNeighbour.x .. " " .. openNeighbour.y)
						--print("neighbour g: " .. neighbour.g)
						--print("openNeighbour g: " .. openNeighbour.g)
						if neighbour.g < openNeighbour.g then
							openNeighbour.g = neighbour.g
							openNeighbour.parent = neighbour.parent
							--print("set parent [" .. openNeighbour.parent.x .. ":" .. openNeighbour.parent.y .. "] = ["  .. neighbour.parent.x .. ":" .. neighbour.parent.y .. "]")
						end
					end
				end
			end
			
		end
		
	end
	
	return false

end


function Astar.isNodeInList(node, list)
	for _, n in pairs(list) do
		if (n.x == node.x) and (n.y == node.y) then
			return true
		end
	end
	return false
end


function Astar.heuristic(start, goal)
	-- manhattan
	return math.abs(goal.x - start.x) + math.abs(goal.y - start.y)
end


function Astar.getElementWithLowestF()
	local bestF, node = Astar.map.width + Astar.map.height, nil
	for i = 1, #Astar.open do
		if Astar.open[i].f < bestF then
			bestF = Astar.open[i].f
			node = Astar.open[i]
		end
	end
	
	return node
end


function Astar.isNodeEqual(node1, node2)
	return (node1.x == node2.x) and (node1.y == node2.y)
end


function Astar.removeNode(_table, node)
	for i = #_table, 1, -1 do
		if Astar.isNodeEqual(_table[i], node) then
			--print("a* remove node from open table: " .. node.x .. " " .. node.y)
			table.remove(_table, i)
			break
		end
	end
end

--[[
map.blockingData should be a 2D data structure:
[0,0][1,0][2,0]...
[0,1][1,1][2,1]
[0,2][1,2][2,2]
...            [x,y]
where 0 = not blocking and > 0 can be blocking depending on the movement type
--]]


function Astar.setObstacles(_obstacles)	-- table containing blocking mapObject types
	Astar.obstacles = _obstacles
end


function Astar.getNeighbours(node)
	local neighbours = {}
	local width, height = Astar.map.width, Astar.map.height
	--local nodeIndex = (node.x + node.y * width) + 1	-- old
	
	-- top neighbour
	if node.y > 1 then
		local neighbourNode = {x = node.x, y = node.y - 1}
		--if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] == 0 then	-- 0 = not blocking
		if not tools.isValueInTable(Astar.map.blockingData[neighbourNode.y][neighbourNode.x], Astar.obstacles) then
			neighbourNode.g = node.g + 1
			neighbourNode.parent = node
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- bottom neighbour
	if node.y < height then
		local neighbourNode = {x = node.x, y = node.y + 1}
		if not tools.isValueInTable(Astar.map.blockingData[neighbourNode.y][neighbourNode.x], Astar.obstacles) then
			neighbourNode.g = node.g + 1
			neighbourNode.parent = node
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- left neighbour
	if node.x > 1 then
		local neighbourNode = {x = node.x - 1, y = node.y}
		if not tools.isValueInTable(Astar.map.blockingData[neighbourNode.y][neighbourNode.x], Astar.obstacles) then
			neighbourNode.g = node.g + 1
			neighbourNode.parent = node
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- right neighbour
	if node.x < width then
		local neighbourNode = {x = node.x + 1, y = node.y}
		if not tools.isValueInTable(Astar.map.blockingData[neighbourNode.y][neighbourNode.x], Astar.obstacles) then
			neighbourNode.g = node.g + 1
			neighbourNode.parent = node
			table.insert(neighbours, neighbourNode)
		end
	end
	
	return neighbours
end


function Astar.constructPath(node)
	local path = {node}
	while node.parent ~= nil do
		--print("path node: " .. node.x .. " : " .. node.y)
		node = node.parent
		table.insert(path, node)
	end
	return path
end


return Astar


--[[
-- a star pathfinding algorithm

local Astar = {}

function Astar.run(map, start, goal)

	Astar.nodes		= {}
	Astar.map = map
	
	local nid = 0
	for h = 1, Astar.map.height do
		for w = 1, Astar.map.width do
			nid = nid + 1
			table.insert(Astar.nodes, {id = nid, x = w, y = h, parentID = nil, g = 0, f = 0})
		end
	end
	
	-- both holding IDs only!
	Astar.open		= {}
	Astar.closed	= {}
	
	local startNode = Astar.getNodeByPosition(start.x, start.y)
	table.insert(Astar.open, startNode.id)
	
	local goalNode = Astar.getNodeByPosition(goal.x, goal.y)
	startNode.g = 0
	startNode.f = startNode.g + Astar.heuristic(startNode, goalNode)
	
	
	while #Astar.open > 0 do
	
		print("a* open nodes: " .. #Astar.open)
		local current = Astar.getElementWithLowestF()
		print("a* current node: " .. current.x .. " " .. current.y)
		if Astar.isNodeEqual(current, goalNode) then	-- same as having the same id
			return Astar.constructPath(goalNode)
		end
		-- otherwise...
		Astar.removeNodeByID(Astar.open, current.id)
		table.insert(Astar.closed, current.id)
		
		for _, neighbour in pairs(Astar.getNeighbours(current)) do
			--neighbour.parent = neighbour.parent or node	--?
			if not Astar.isNodeInList(neighbour, Astar.closed) then
				neighbour.f = neighbour.g + Astar.heuristic(neighbour, goalNode)
				
				if not Astar.isNodeInList(neighbour, Astar.open) then
					print("insert node " .. neighbour.x .. " " .. neighbour.y .. " into open list")
					table.insert(Astar.open, neighbour.id)
				else
					print("node " .. neighbour.x .. " " .. neighbour.y .. " is already in open list")
					for _, openNeighbourID in pairs(Astar.open) do
						local openNeighbour = Astar.nodes[openNeighbourID]
						print("check neighbour " .. openNeighbour.x .. " " .. openNeighbour.y)
						print("neighbour g: " .. neighbour.g)
						print("openNeighbour g: " .. openNeighbour.g)
						if neighbour.g < openNeighbour.g then
							openNeighbour.g = neighbour.g
							openNeighbour.parentID = neighbour.parentID
							--print("set parent [" .. openNeighbour.parent.x .. ":" .. openNeighbour.parent.y .. "] = ["  .. neighbour.parent.x .. ":" .. neighbour.parent.y .. "]")
						end
					end
				end
			end
			
		end
		
	end
	
	return false

end


function Astar.getNodeByPosition(_x, _y)
	for i = 1, #Astar.nodes do
		if (Astar.nodes[i].x == _x) and (Astar.nodes[i].y == _y) then
			return Astar.nodes[i]
		end
	end
end


function Astar.isNodeInList(node, list)
	for _, n in pairs(list) do
		if n == node.id then
			return true
		end
	end
	return false
end


function Astar.heuristic(start, goal)
	-- manhattan
	return math.abs(goal.x - start.x) + math.abs(goal.y - start.y)
end


function Astar.getElementWithLowestF()
	local bestF, nodeID = Astar.map.width + Astar.map.height, nil
	for i = 1, #Astar.open do
		if Astar.nodes[Astar.open[i] ].f < bestF then
			bestF = Astar.nodes[Astar.open[i] ].f
			nodeID = Astar.open[i]
		end
	end
	
	return Astar.nodes[nodeID]
end


function Astar.isNodeEqual(node1, node2)
	--return (node1.x == node2.x) and (node1.y == node2.y)
	return node1.id == node2.id
end


function Astar.removeNodeByID(_table, nodeID)
	for i = #_table, 1, -1 do
		if _table[i] == nodeID then
			print("a* remove node from open table: " .. nodeID)
			table.remove(_table, i)
			break
		end
	end
end

--
map.blockingData should be a 2D data structure:
[0,0][1,0][2,0]...
[0,1][1,1][2,1]
[0,2][1,2][2,2]
...            [x,y]
where 0 = not blocking and 1 = blocking
--

function Astar.getNeighbours(node)
	local neighbours = {}
	local width, height = Astar.map.width, Astar.map.height
	--local nodeIndex = (node.x + node.y * width) + 1	-- old
	
	-- top neighbour
	if node.y > 1 then
		local neighbourNode = Astar.getNodeByPosition(node.x, node.y - 1)
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] == 0 then	-- 0 = not blocking
			neighbourNode.g = node.g + 1
			neighbourNode.parentID = node.id
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- bottom neighbour
	if node.y < height then
		local neighbourNode = Astar.getNodeByPosition(node.x, node.y + 1)
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] == 0 then
			neighbourNode.g = node.g + 1
			neighbourNode.parentID = node.id
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- left neighbour
	if node.x > 1 then
		local neighbourNode = Astar.getNodeByPosition(node.x - 1, node.y)
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] == 0 then
			neighbourNode.g = node.g + 1
			neighbourNode.parentID = node.id
			table.insert(neighbours, neighbourNode)
		end
	end
	
	-- right neighbour
	if node.x < width then
		local neighbourNode = Astar.getNodeByPosition(node.x + 1, node.y)
		if Astar.map.blockingData[neighbourNode.y][neighbourNode.x] == 0 then
			neighbourNode.g = node.g + 1
			neighbourNode.parentID = node.id
			table.insert(neighbours, neighbourNode)
		end
	end
	
	return neighbours
end


function Astar.constructPath(node)
	local path = {node}
	while node.parentID ~= nil do
		print("node: " .. node.x .. " : " .. node.y)
		table.insert(path, node)
		node = Astar.nodes[node.parentID]
	end
	return path
end


return Astar

--]]
