-- creature <- movingentity
movingentity = require "scripts/entities/movingentity"
--soundmgr	= require "scripts/soundmgr"

local Creature = {}

function Creature:create(_entity)
	local creature = movingentity:create({x = _entity.x, y = _entity.y})
	
	creature.entityCategory	= "ec_creature"
	
	creature.movementType		= "walk"				-- movement type can be: walk, fly, swim
	creature.moving				= false					-- currently moving?
	
	creature.playerControlled = false	-- deprecated?
	creature.abilities = {}	-- this contains all attack/utility/heal etc abilities; passive + active
	creature.currentTargetID = nil
	creature.lastHurtTime = 0
	
	creature.health = 0
	creature.maxHealth = 0
	creature.shield = 0
	creature.maxShield = 0
	creature.displayHealthBar = true
	creature.displayShieldBar = false
	
	creature.partSys = {}
	creature.hurtParticleImage = love.graphics.newImage("assets/graphics/player1.png")
	creature.hurtParticleTimer = 0
	
	creature.sounds = {}

	
	function creature:setAttackTarget(targetID)
		self.currentTargetID = targetID
	end
	
	
	function creature:isAlive()
		return (self.health > 0)
	end
	
	
	-- moving etc
	function creature:move(dt)
		local currSpeed = 0
		local colliding = false
		local point
		local collisionRect
		
		-- environment contains obstacles based on the creatures movement type
		if self.environment ~= nil then
			-- table that contains rectangles of blocking mapObjects around the moving entity
			for k, rect in pairs(self.environment) do
				--print("creature env: rect " .. rect.x .. ":" .. rect.y)
				local px, py = self:precalculatePosition(dt)
				local circle = {radius = self.size / 2, x = px, y = py}
				point = tools.getCollisionPoint(circle, rect)
				if point ~= nil then
					-- player is about to collide with the mapObject in the next step -> do not execute the movement!
					colliding = true
					collisionRect = rect
					--print("creature collision: " .. rect.x .. " " .. rect.y)
					break
				end
			end
		
		end
		
		
		
		-- moving at all?
		if self.movingVector.x ~= 0 or self.movingVector.y ~= 0 then
			self.moving = true
			currSpeed = tools.vectorGetLength(self.movingVector)
			--print("movement speed: " .. math.floor(currSpeed))
			
			-- trim vector to assure a maximum speed
			if currSpeed > self.maxSpeed then
				self.movingVector = tools.trimVector(self.movingVector, self.maxSpeed)
			end
			
			-- the actual movement shall only be executed if the collision check is ok
			if not colliding then
				self.x = self.x + self.movingVector.x * dt
				self.y = self.y + self.movingVector.y * dt
			else
				--print("do not move because collisiong comming!")
				
				-- now let the entity react to the collision
				if self.collideAtPoint ~= nil then
					self:collideAtPoint(point, collisionRect)
				end
			end
			
			--self.rotation = math.atan2(self.movingVector.y, self.movingVector.x) + math.pi * 0.5
			
			if math.abs(self.movingVector.x) < 1 then
				-- prevent minor movement with very low directional speed
				self.movingVector.x = 0
			else
				-- diminish directional movement
				self.movingVector.x = self.movingVector.x * self.inertnessFactor
			end
			
			if math.abs(self.movingVector.y) < 1 then
				-- prevent minor movement with very low directional speed
				self.movingVector.y = 0
			else
				-- diminish directional movement
				self.movingVector.y = self.movingVector.y * self.inertnessFactor
			end
			
			-- rotate the creature
			if (self.rotateByMovement) and (currSpeed > 3) then	-- do not rotate when the creature is only barely moving
				self.rotation = math.atan2(self.movingVector.y, self.movingVector.x) + math.pi * 0.5	-- add pi/2 because the atan2 refers to the x-axis while the rotation refers to the y-axis as 0
			end
			
		else	-- not moving
		
			self.moving = false
			
		end
		
		-- update speed
		self.speed = currSpeed
	end
	
	
	function creature:updateParticleSystems(dt)
	
		-- update particle systems
		local numberOfParticleSystems = #self.partSys
		if numberOfParticleSystems > 0 then
			for i = numberOfParticleSystems, 1, -1 do
				local pMin, pMax = self.partSys[i].system:getParticleLifetime()
				local lifetime = self.partSys[i].system:getEmitterLifetime()
				-- also take the particle lifetime into consideration, not only the emitter's!
				-- emitter lifetime of -1 means it emitts forever -> do not remove
				if (self.partSys[i].startTime + lifetime + pMax <= GAMETIME) and (lifetime > -1) then
					--print("ps: remove")
					table.remove(self.partSys, i)
				else
					--if self.partSys[i].name == "run" then
						--local x,y = self.partSys[i].system:getPosition()
						--print(x,y)
						--self.partSys[i].system:setPosition(self.x + offsetX, self.y + offsetY)
						--self.partSys[i].pos = {X = self.x + offsetX, Y = self.y + offsetY}
					--end
					self.partSys[i].system:update(dt)
				end
			end
		end
		
	end
	
	
	function creature:isDamageLethal(_damage)
		return (self.shield + self.health) <= _damage
	end
	
	
	-- TODO: if nonPlayer creature: if it is not yet attacking ANY target, then attack the attacker
	function creature:hurt(_damage, _sourceID, _ignoreshield)
		--print("creature hurt: " .. _damage)
		local offsetX, offsetY = self.imageWidth / 2, self.imageHeight / 2
		
		-- 
		local damage = _damage or 1
		local ignoreshield = _ignoreshield or false
		local damageHealth, damageShield = 0, 0
		if self.shield == 0 or ignoreshield then
			-- creature has no shield OR takes true damage -> take full damage to healthbar
			damageHealth = damage
		else
			-- creature has shield, damage shield first and raw health if necessary
			damageShield = math.min(damage, self.shield)
			damageHealth = damage - damageShield
		end
		self.shield = math.max(0, self.shield - damageShield)
		self.health = math.max(0, self.health - damageHealth)
		--print("new health: " .. self.health)
		
		self.lastHurtTime = GAMETIME
		
		if self.health == 0 then
			self:destroy()
		end
		
		if (self.entityType ~= "player") and (_sourceID ~= nil) then
			--print("creature: set attack target ID")
			creature:setAttackTarget(_sourceID)
		end
		
		-- TODO: display particle color based on shield / raw health loss - adjust colors!
		-- disply particles
		local r, g, b = 0, 0, 255
		if GAMETIME - self.hurtParticleTimer > 0.5 then
			local sys = {}
			if damageHealth > 0 then
				r, g, b = 255, 0, 0
			end
			sys.system = love.graphics.newParticleSystem(self.hurtParticleImage, 64)
			sys.system:setParticleLifetime(0.5, 1) -- Particles live at least (min)s and at most (max)s.
			sys.system:setEmissionRate(64)
			sys.system:setSizeVariation(1)
			sys.system:setEmitterLifetime(0.1)
			sys.system:setEmissionArea("uniform", self.size / 2, self.size / 2)
			sys.system:setLinearAcceleration(-20, -20, 20, 20) -- Random movement in all directions.
			sys.system:setColors(r, g, b, 255, r, g, b, 0) -- Fade to transparency.
			sys.system:setSizes(0.1, 0.2)
			sys.pos = {X = self.x + offsetX, Y = self.y + offsetY}
			sys.startTime = GAMETIME
			sys.name = ""
			table.insert(self.partSys, sys)
			
			self.hurtParticleTimer = GAMETIME
		end
		
		-- play sound if possible
		if self.sounds["hurt"] ~= nil then
			local rnd = math.random(1, #self.sounds["hurt"])
			local sound = self.sounds["hurt"][rnd]
			soundmgr.playSound(sound)
		end
		
	end
	
	
	function creature:collideAtPoint(point, rect)	--dt
		--self.colliding = true
		local px, py = self:getPosition()
		local currSpeed = tools.vectorGetLength(self.movingVector)
		--local centerP = {x = px, y = py}
		--local centerR = {x = rect.x + rect.width / 2, y = rect.y + rect.height / 2}
		--self.collisionLine = {centerP, centerR}	--tools.vectorSub(centerP, centerR)
		self.collisionPoint = point
		
		-- 43
		-- 12
		local corner = tools.getCorner(point, rect)
		--[[
		if corner > 0 then
			print("hitting corner " .. corner)
			local speed_dx, speed_dy = math.abs(self.movingVector.x), math.abs(self.movingVector.y)
			if speed_dy < 20 then
				if corner < 3 then					-- hitting from below
					self.movingVector.y = - 10
				else								-- hitting from above
					self.movingVector.y = 10
				end
			end
			
			if speed_dx < 20 then
				if corner == 1 or corner == 4 then	-- hitting left side
					self.movingVector.x = - 10
				else								-- hitting right side
					self.movingVector.x = 10
				end
			end

		end
		--]]
		
		local diffX, diffY = math.abs(px - point.x), math.abs(py - point.y)
		if diffX > (self.size / 2) then
			self.movingVector.x = - self.movingVector.x * 0.02
			--self.movingVector.y = self.movingVector.y * 10
		elseif diffY > (self.size / 2) then
			self.movingVector.y = - self.movingVector.y * 0.02
			--self.movingVector.x = self.movingVector.x * 10
		else
			--[[
			self.movingVector.x = - self.movingVector.x
			self.movingVector.y = - self.movingVector.y
			--]]
			if corner > 0 then
				--print("hitting corner " .. corner)
				self.movingVector = tools.vectorAdd(self.movingVector, {x = px - point.x, y = py - point.y})
				--self:accelerate(AVG_DT, px - point.x, py - point.y)

			end
			
		end
	
	end
	
	
	function creature:draw()
	
		--[[
		local offsetX, offsetY = self.size / 2, self.size / 2
		if self.image == nil then
			love.graphics.draw(self.images[self.currentImageIndex], self.x + offsetX, self.y + offsetY, self.rotation, self.imageScale, self.imageScale, offsetX / self.imageScale, offsetY / self.imageScale)
		else
			love.graphics.draw(self.image, self.x + offsetX, self.y + offsetY, self.rotation, self.imageScale, self.imageScale, offsetX / self.imageScale, offsetY / self.imageScale)
		end
		--]]
		
		-- simple health bar
		if self.displayHealthBar then
			local health = self.health / self.maxHealth
			love.graphics.setColor(0, 0, 0, 0.9)
			--local height = self.health / 33
			local height = 6
			--love.graphics.rectangle("fill", self.x, self.y - height + 1, self.size, height)
			love.graphics.rectangle("fill", self.x, self.y - height + 1, self.size, height)
			local r, g, b = tools.getRGBbyHealth(health)
			love.graphics.setColor(r, g, b, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - height + 2, (self.size - 2) * health, height - 2)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		-- simple shield bar
		if self.displayShieldBar and self.shield > 0 then
			local shield = self.shield / self.maxShield
			love.graphics.setColor(0, 0, 0, 0.9)
			love.graphics.rectangle("fill", self.x, self.y - 12, self.size, 6)
			love.graphics.setColor(0.8, 0.8, 0.8, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - 11, (self.size - 2) * shield, 4)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		-- particle systems
		local numberOfParticleSystems = #self.partSys
		if numberOfParticleSystems > 0 then
			for i = 1, numberOfParticleSystems do
				--local x,y = self.partSys[i].system:getPosition()
				love.graphics.draw(self.partSys[i].system, self.partSys[i].pos.X, self.partSys[i].pos.Y)
			end
		end
		
	end
	
	
	return creature
end


return Creature
