-- moving entity <- entity

entity = require "scripts/entity"

local MovingEntity = {}

function MovingEntity:create(_entity)
	local entity = entity:create({id = _entity.id, x = _entity.x, y = _entity.y})

	print("creating moving entity named ".. _entity.id .." at " .. (_entity.x or 0) .. ";" .. (_entity.y or 0))
	
	entity.movingVector		= {x = 0, y = 0}		-- 2D directional vector
	entity.speed			= 0						-- current speed
	entity.maxSpeed			= 200					-- (world units / pixels?)
	entity.rotation			= 0
	entity.maxAcceleration 	= 15
	entity.inertnessFactor	= 0.9					-- how fast the entity will slow down
	entity.movementType		= "walk"				-- movement type can be: walk, fly, swim
	entity.moving			= false					-- currently moving?
	entity.rotateByMovement = true

	-- generic functions
	function entity:load() end
	function entity:tick(dt) end
	function entity:draw() end

	
	-- get where the entity is expected to be in the next step; top left corner
	function entity:precalculateXandY(dt)
		return self.x + self.movingVector.x * dt, self.y + self.movingVector.y * dt
	end
	
	
	-- get where the entity is expected to be in the next step; center
	-- use dx and dy to give a different movingVector
	function entity:precalculatePosition(dt, dx, dy)
		local offsetX, offsetY = (self.size or self.width) / 2, (self.size or self.height) / 2
		return self.x + (dx or self.movingVector.x) * dt + offsetX, self.y + (dy or self.movingVector.y) * dt + offsetY
	end
	
	
	function entity:accelerate(dt, dx, dy)
		local colliding = false
		
		if self.environment ~= nil then
			-- table that contains rectangles of blocking mapObjects around the moving entity
			for k, rect in pairs(self.environment) do
				local px, py = self:precalculatePosition(dt, dx, dy)
				local circle = {radius = self.size / 2, x = px, y = py}
				point = tools.getCollisionPoint(circle, rect)
				if point ~= nil then
					-- player is about to collide with the mapObject in the next step -> ignore the acceleration in this direction!
					colliding = true
					break
				end
			end
		end
		
		if not colliding then	--and self.collisionPoint == nil
			self.movingVector.x = self.movingVector.x + dx
			self.movingVector.y = self.movingVector.y + dy
		end
		
	end
	
	
	-- moving etc
	function entity:move(dt)
		local currSpeed = 0
		local colliding = false
		local point
		local collisionRect
		
		if self.environment ~= nil then
			-- table that contains rectangles of blocking mapObjects around the moving entity
			for k, rect in pairs(self.environment) do
				local px, py = self:precalculatePosition(dt)
				local circle = {radius = self.size / 2, x = px, y = py}
				point = tools.getCollisionPoint(circle, rect)
				if point ~= nil then
					-- player is about to collide with the mapObject in the next step -> do not execute the movement!
					colliding = true
					collisionRect = rect
					--print("collision: " .. rect.x .. " " .. rect.y)
					break
				end
			end
		
		end
		
		
		
		-- moving at all?
		if self.movingVector.x ~= 0 or self.movingVector.y ~= 0 then
			self.moving = true
			currSpeed = tools.vectorGetLength(self.movingVector)
			--print("movement speed: " .. math.floor(currSpeed))
			
			-- trim vector to assure a maximum speed
			if currSpeed > self.maxSpeed then
				self.movingVector = tools.trimVector(self.movingVector, self.maxSpeed)
			end
			
			-- the actual movement shall only be executed if the collision check is ok
			if not colliding then
				self.x = self.x + self.movingVector.x * dt
				self.y = self.y + self.movingVector.y * dt
			else
				--print("do not move because collisiong comming!")
				
				-- now let the entity react to the collision
				if self.collideAtPoint ~= nil then
					self:collideAtPoint(point, collisionRect)
				end
			end
			
			--self.rotation = math.atan2(self.movingVector.y, self.movingVector.x) + math.pi * 0.5
			
			if math.abs(self.movingVector.x) < 1 then
				-- prevent minor movement with very low directional speed
				self.movingVector.x = 0
			else
				-- diminish directional movement
				self.movingVector.x = self.movingVector.x * self.inertnessFactor
			end
			
			if math.abs(self.movingVector.y) < 1 then
				-- prevent minor movement with very low directional speed
				self.movingVector.y = 0
			else
				-- diminish directional movement
				self.movingVector.y = self.movingVector.y * self.inertnessFactor
			end
			
			-- rotate the player
			if (self.rotateByMovement) and (currSpeed > 3) then	-- do not rotate when the player is only barely moving
				self.rotation = math.atan2(self.movingVector.y, self.movingVector.x) + math.pi * 0.5	-- add pi/2 because the atan2 refers to the x-axis while the rotation refers to the y-axis
			end
			
		else	-- not moving
		
			self.moving = false
			
		end
		
		-- update speed
		self.speed = currSpeed
	end

	
	return entity
end


return MovingEntity