local propertymgr = require "scripts/propertymgr"

-- tools

local Tools = {}

function Tools.isPointInsideRect(px, py, rect)
	
	if px > rect.x and
		px < rect.x + rect.width and
		py > rect.y and
		py < rect.y + rect.height then
		
		return true
	end
	
	return false

end

function Tools.areRectsColliding(rect_1, rect_2)

	if rect_1.x + rect_1.width > rect_2.x and
	   rect_1.x < rect_2.x+rect_2.width   and
	   rect_1.y + rect_1.height > rect_2.y and
	   rect_1.y < rect_2.y+rect_2.height   then
	
	   	return true
	end

	return false
end


function Tools.getVectorFromAngle(_rotation)
	return {x = math.cos(_rotation), y = math.sin(_rotation)}
end


function Tools.getSign(number)
	if number < 0 then
		return -1
	else
		return 1
	end
end


function Tools.clamp(value, min, max)
	return math.max(math.min(value, max), min)
end

function Tools.areCirclesColliding(circle1, circle2)
	local dist = Tools.getDistance({x = circle1.x, y = circle1.y}, {x = circle2.x, y = circle2.y})
	--return circle1.radius + circle2.radius <= Tools.getDistance(circle1.pos, circle2.pos)	-- wrong ...
	return circle1.radius * circle1.radius + circle2.radius * circle2.radius >= Tools.getDistanceSquared({x = circle1.x, y = circle1.y}, {x = circle2.x, y = circle2.y})
	--return circle1.radius + circle2.radius >= dist
end


-- TODO: right now assuming we have a circle and a rectangle
function Tools.getCollisionPoint(circle, rect)
	if Tools.areCircleAndRectColliding(circle, rect) then
		local point = {x = Tools.clamp(circle.x, rect.x, rect.x + rect.width), y = Tools.clamp(circle.y, rect.y, rect.y + rect.height)}
		return point
	else
		return nil
	end
end


function Tools.areCircleAndRectColliding(circle, rect)

	local rx, ry, dx, dy
	rx = Tools.clamp(circle.x, rect.x, rect.x + rect.width)
	ry = Tools.clamp(circle.y, rect.y, rect.y + rect.height)
	dx = circle.x - rx
	dy = circle.y - ry
	return (dx * dx + dy * dy) < (circle.radius * circle.radius)

end

-- TODO: build some detection if object is circle or rectangle by their properties
function Tools.areObjectsColliding(object_1, object_2)
--[[
	if isCircle(object_1) and isCircle(object_2) then
		return Tools.areCirclesColliding(object_1, object_2)
	end
--]]
end


function Tools.getDistance(pos1, pos2)

	return math.sqrt((pos2.x - pos1.x) * (pos2.x - pos1.x) + (pos2.y - pos1.y) * (pos2.y - pos1.y))
	
end

-- a little less handy but faster as it does not require a root
function Tools.getDistanceSquared(pos1, pos2)

	return (pos2.x - pos1.x) * (pos2.x - pos1.x) + (pos2.y - pos1.y) * (pos2.y - pos1.y)
	
end


function Tools.vectorGetLength(vector)

	return math.sqrt(vector.x * vector.x + vector.y * vector.y)
	
end


function Tools.isCorner(point, rect)
	return (point.x == rect.x) and (point.y == rect.y)
		or (point.x == rect.x + rect.width) and (point.y == rect.y)
		or (point.x == rect.x + rect.width) and (point.y == rect.y + rect.height)
		or (point.x == rect.x) and (point.y == rect.y + rect.height)
end

-- 43
-- 12
function Tools.getCorner(point, rect)
	local corner = 0
	if (point.x == rect.x) and (point.y == rect.y + rect.height) then
		corner = 1
	elseif (point.x == rect.x + rect.width) and (point.y == rect.y + rect.height) then
		corner = 2
	elseif (point.x == rect.x + rect.width) and (point.y == rect.y) then
		corner = 3
	elseif (point.x == rect.x) and (point.y == rect.y) then
		corner = 4
	end
	
	return corner
end


function Tools.setVectorLength(vector, length)
	local newLen = length or 1
	local currLen = Tools.vectorGetLength(vector)
	local vx, vy
	vx = math.floor(vector.x * newLen/currLen)
	vy = math.floor(vector.y * newLen/currLen)
	return {x = vx, y = vy}
end


function Tools.trimVector(vector, maximumLength)

	local maxLen = maximumLength or 1
	local currLen = Tools.vectorGetLength(vector)
	
	if currLen < maxLen then
		-- vector already short enough
		return vector
	else
		-- vector norm too big
		local vx, vy
		vx = math.floor(vector.x * maxLen/currLen)
		vy = math.floor(vector.y * maxLen/currLen)
		return {x = vx, y = vy}
	end
	
end


function Tools.vectorSub(vec_1, vec_2)

	return {x = vec_2.x - vec_1.x, y = vec_2.y - vec_1.y}
end


function Tools.vectorAdd(vec_1, vec_2)

	return {x = vec_2.x + vec_1.x, y = vec_2.y + vec_1.y}
end


-- returns true if a given value can be found inside the table, false otherwise
function Tools.isValueInTable(_value, _table)
	for index, value in ipairs(_table) do
		if value == _value then
			return true
		end
	end

	return false
end


-- returns whether the given map object type is blocking or not (better to look it up in a table via its type then having to handle the object itself)
function Tools.isMapObjectTypeBlocking(_type)

	if Tools.isValueInTable(_type, propertymgr.getBlockingMapObjects()) then
		return true
	else
		return false
	end
	
end


-- returns whether the given map object type is blocking or not; based on the creatures movementType
function Tools.isMapObjectTypeBlockingByMovementType(_type, _movementType)
	if Tools.isValueInTable(_type, propertymgr.getObstaclesByMovementType(_movementType)) then
		return true
	else
		return false
	end
end


function Tools.getObstaclesByMovementType(_movementType)
	return propertymgr.getObstaclesByMovementType(_movementType)
end


function Tools.round(value)
	return math.floor(value + 0.5)
end


-- get RGB relative to given health: turn from green to yellow to red (form 100% to 0% health)
-- TODO: update this in settlers!(?)
function Tools.getRGBbyHealth(health)
	local R, G, B
	--[[
	if health == 100 then
		R, G, B = 0, 255, 0
	elseif health > 50 then
		R = 255 - health * 2.55
		G = 255
		B = 0
	else
		R = 255
		G = health * 5.1
		B = 0
	end
	return R/255, G/255, B/255
	--]]
	
	if health > 0.5 then
		R = 1 - health + 0.5
		G = 1
		B = 0
	else
		R = 1
		G = health * 2
		B = 0
	end
	return R, G, B
end

-- note: two separators next to each other = no table entry (careful when passing "empty" parameters)
function Tools.string_split(strng, sep)
	-- see: https://stackoverflow.com/questions/1426954/split-string-in-lua
	local t = {}
	local i = 1
	for str in string.gmatch(strng, "([^"..sep.."]+)") do
		t[i] = str
		i = i + 1
	end
	return t
end



return Tools
